/*
    v0.01 - original
    v0.02 - minor bugs
            added tsrcommander
            added more error reporting
            fixed big leak in exTAP_EPG_GetSize
            fixed bug in hook code causing load order crashes
            hacky support for non-GMT timezones

    a tap to enhance the internal EPG
    RKEY_Recall, bring up menu
    or if tsrcommander is running; use RKEY_Exit

on startup
    *-  increase size of internal EPG on tap startup
    *-  load internal EPG from disk on tap load
    *-  enable dsmcc capture if tuned to 4tv and betweeen 3-4am

on shutdown
    *-  save internal EPG to disk on power off

on demand
    *-  load internal EPG from disk
    *-  save internal EPG to disk
    *-  export internal EPG to mei
    *-  enable/disable dsmcc capture of current channel
    *-  display number of internal EPG events
    *-  exit tap
    *-  enable/disable conversion broadcast eit data
    *-  delete entire internal EPG

on EIT data
    *-  ignore out of date events
    *-  convert content_descriptors to extended_descriptors
    *-  convert component_descriptors to extended_descriptors

every 10 minutes
    *-  enable dsmcc capture if tuned to 4tv and betweeen 3-4am

on channel change
    *-  disable dsmcc capture

tbd later ...
    *-  config options
    *-  purge by channel/date
    *-  filter by channel/date
    *-  delete out of date events
    *-  import/export eit
    *-  import/export mei
    *-  import/export tgd

    *-  process 4tv dsmcc data
    *-  support for >5000 events
    *-  etc.
*/
#include <tap.h>
#include "exTap.h"
#include "eit.h"

//#define KEY_REMAPPING         //enable use of the arrow keys, list


//Every TAP application needs unique ID and TAP allocations that have same ID cant execute at same time.
//ID consists of DWORD( 4byte ) and the MSB bit should be set by 1.
//The following is how to use in source codes.
#define ID_USERTAP 0xbdb00502
TAP_ID( ID_USERTAP );

//Every TAP application has to include Program name, Author name, and Program description as following
//manners in source code.
TAP_PROGRAM_NAME("epg_extender v0.02");
TAP_AUTHOR_NAME("bdb");
TAP_DESCRIPTION("a tap to enhance the internal EPG");
TAP_ETCINFO(__DATE__);

#include "TSRCommander.inc"

//----------------------------------------------------------------------------
//menu stuff
typedef struct {
    int  id;
    char str[0x20];
}type_item;

typedef struct {
    int  id;
    int  visible;
    TYPE_Window win;
    int  rgn;
    char *fixme_titlep;
    type_item item[0x20];
}type_menu;

extern dword message_showing;
extern dword message_rgn;
extern dword ShowMessage(char* lpMessage, char* lpMessage1, dword dwDelay, int wait);
extern void ShowMenu(type_menu *menu);
extern void DeleteMenu(type_menu *menu);

//----------------------------------------------------------------------------
#define FW_MAX_EVENTS      5000

typedef struct{
    int valid;
    int table_id;
    int section_length;
    int table_id_extension;
    int version_number;
    int current_next_indicator;
    int section_number;
    int last_section_number;
    int event_id;
    int count;
    int len;
} type_table;

#define MAX_TABLE_ID    0x1000
int num_tables;
type_table *table;

int dsmcc_enabled = 0;
int eit_enabled = 0;
int epg_enabled = 0;

//----------------------------------------------------------------------------
type_menu main_menu = {
    0,
    0,
    {0},
    0,
    (char *)__tap_program_name__,
    {
        {7, "xxx",                 },
        {0, "Load epg.dat",         },
        {1, "Save epg.dat",         },
        {2, "Export MyStuff.mei",   },
        {3, "Enable DSMCC Hook",    },
        {4, "Show Num Events",      },
        {5, "Exit Tap",             },
        {6, "More ...",             },
    },
};
type_menu more_menu = {
    0,
    0,
    {0},
    0,
    (char *)__tap_program_name__,
    {
        {6, "xxx",                  },

        {0, "Apply EIT Hook",       },
        {1, "Delete all Events",    },
        {2, "Disable EPG",          },
        {3, "Test1 (mem info)",     },
        {4, "Test2 (shutdown)",     },
        {5, "Test3 (dump channel info)",  },
    },
};

enum {
    STATE_IDLE = 0,
    STATE_LOAD_EPG_BUSY,
    STATE_LOAD_EPG_CLOSE,
};
int idle_state = STATE_IDLE;
int ten_minutes = 0;

int TAP_Main(void);
dword TAP_EventHandler(word event, dword param1, dword param2);
int main_menu_handler(type_menu *menu, dword param1);
int more_menu_handler(type_menu *menu, dword param1);
int key_handler(dword param1);

int is_tsr_commander_running(void);
void dump(dword addr, int length);
int dump_channels(void);
int get_channel_number(int service_id);

int init(void);
int close(void);

int shutdown_hook_id = -2;
int shutdown = 0;
int apply_shutdown_hook();
int remove_shutdown_hook();
dword shutdown_hook(dword param1, dword param2);

int dsmcc_hook_id = -2;
TYPE_File *dsmcc_Op = 0;
int apply_dsmcc_hook();
int remove_dsmcc_hook();
dword dsmcc_hook(dword param1, dword param2);
int section_process_3b3c(byte *byte, int len);

int eit_hook_id = -2;
int apply_eit_hook();
int remove_eit_hook();
dword eit_hook(dword param1, dword param2);

//void *service_hash = 0;
//int init_service_id_filter();
//int close_service_id_filter();
dword service_id_filter(byte *buf);

int load_epg(char *filename);
int save_epg(char *filename);
int get_genre(type_eventtable *et, char **value);
int export_mei(char *filename);

//----------------------------------------------------------------------------
enum {
    DELAY_SHORT  = 100,     //1s
    DELAY_MEDIUM = 500,     //5s
    DELAY_LONG   = 1000,    //10s
};
enum {
    LOG_SERIAL = 1,
    LOG_SCREEN_CONT  = 2,
    LOG_SCREEN_WAIT  = 4,
    LOG_FILE = 8,
};
char strLog[0x100];
TYPE_File *OpLog = 0;
int log_message(int log, char *str, int delay);

//----------------------------------------------------------------------------
//tsr commander stuff
bool TSRCommanderExitTAP()
{
    close();
    TAP_Exit();
    return FALSE;
}

void TSRCommanderConfigDialog()
{
    ShowMenu(&main_menu);
}

//----------------------------------------------------------------------------
int TAP_Main(void)
{
    word SysID = *(word*)pSysID;
    int ret;

    sprintf(strLog, "Opening");
    log_message(LOG_SERIAL, strLog, DELAY_SHORT);
    ret = exTAP_Init();
    if (ret) {
        sprintf(strLog, "Unsupported platform - %d:%04x", *(word*)pSysID, _appl_version);
        log_message(LOG_SCREEN_WAIT|LOG_SERIAL, strLog, DELAY_LONG);
        return 0;
    }
    sprintf(strLog, "Opening");
    log_message(LOG_SCREEN_CONT|LOG_SERIAL, strLog, DELAY_SHORT);

    init();
    TSRCommanderInit(0, TRUE);

    return 1;                 //0=application mode, 1=tsr mode
}

//----------------------------------------------------------------------------
dword TAP_EventHandler(word event, dword param1, dword param2)
{
    TYPE_TapChInfo channel_info;
    int channel_type, channel_number;
#ifdef KEY_REMAPPING
    int key_patched;
#endif
    int ret;
    dword state, subState;

    switch (event) {
        case EVT_IDLE:
//            TAP_Print("\nEVT_IDLE, event=%x, param1=%x, param2=%x", event, param1, param2);
            if (message_showing) {
                if (TAP_GetTick() > message_showing) {
                    TAP_Osd_Delete(message_rgn);    // release rgn-handle
                    message_showing = 0;
                    message_rgn = 0;
                }
            }
            if (ten_minutes) {
                if (TAP_GetTick() > ten_minutes) {
                    ten_minute_update();
                    ten_minutes = TAP_GetTick() + (10*60*100);
                }
            }
            switch (idle_state) {
                case STATE_IDLE:
                    break;
                case STATE_LOAD_EPG_BUSY:
                case STATE_LOAD_EPG_CLOSE:
                    load_epg(NULL);
                    break;
            }
    		TSRCommanderWork();
            break;

        case EVT_KEY:
//            TAP_Print("\nEVT_KEY, event=%x, param1=%x, param2=%x", event, param1, param2);
#ifdef KEY_REMAPPING
            key_patched = ((param2 > 0x80000000) && (param2 < 0x84000000)) ? 1 : 0;
            if (!key_patched) {
                sprintf(strLog,"ERROR: Key Mapping patch not applied, aborting");
                log_message(LOG_SCREEN_WAIT|LOG_SERIAL, strLog, DELAY_LONG);
                close();
                TAP_Exit();
                return param1;
            }
            if (!param1) return param1;
            //fix the key code
            param1 = (*(dword *)param2) ? *(dword *)param2 : param1;
//            TAP_Print(", *param2=%x", param1);
#endif

///*
//master shutdown tap only
            //always process RKEY_Power, even if showing another meneu
            if (param1 == RKEY_Power) {
                param1 = key_handler(param1);
            }
            else {
//*/
                TAP_GetState(&state, &subState);
                if ((state == STATE_Normal) && (subState == SUBSTATE_Normal)) {
                    if (main_menu.visible) {
                        param1 = main_menu_handler(&main_menu, param1);
                    }
                    else
                    if (more_menu.visible) {
                        param1 = more_menu_handler(&more_menu, param1);
                    }
                    else {
                        param1 = key_handler(param1);
                    }
                }
///*
            }
//*/
#ifdef KEY_REMAPPING
            //*param2 needs an unmapped code
            *(dword *)param2 = param1;
            //param1 needs a mapped code
            switch(param1) {
                case exRKEY_TvRadio:
                    param1 = RKEY_TvRadio;
                    break;
                case exRKEY_Exit:
                    param1 = RKEY_Exit;
                    break;
                case exRKEY_VolUp:
                    param1 = RKEY_VolUp;
                    break;
                case exRKEY_VolDown:
                    param1 = RKEY_VolDown;
                    break;
                case exRKEY_ChUp:
                    param1 = RKEY_ChUp;
                    break;
                case exRKEY_ChDown:
                    param1 = RKEY_ChDown;
                    break;
            }
#endif
            break;
        case EVT_SVCSTATUS:
            //... not as per user guide
            //d[31]=0
            //d[28]=0 - sub, d[28]=1 - main
            //d[24]=0 - ok,  d[24]=1 - error ?
            //d[16]=0 - video, d[16]=1 - audio
            //d[13:0] - pid
            //check if main channel has switched
            //only switches if audio/video changes, not if dsmcc ??
            //sub channel behaves differently from main channel!
            //its pants
            if (param1 & 0x10000000) {
                TAP_Channel_GetCurrent(&channel_type, &channel_number);
                TAP_Channel_GetInfo(channel_type, channel_number, &channel_info);
                //is it 4tv
                if (channel_info.svcId != 0x6980) {
                    if (dsmcc_enabled) {
                        sprintf(strLog, "Removing DSMCC Hook");
                        log_message(LOG_SCREEN_CONT|LOG_SERIAL, strLog, DELAY_SHORT);
                        remove_dsmcc_hook();
                        break;
                    }
                }
            }
            break;
        case EVT_RBACK:
            break;
        case EVT_UART:
            break;
        case EVT_VIDEO_CONV:
            break;
        case EVT_HOOK:
            if (param1 == shutdown_hook_id) param1 = shutdown_hook(param1, param2);
            if (param1 == eit_hook_id) param1 = eit_hook(param1, param2);
            if (param1 == dsmcc_hook_id) param1 = dsmcc_hook(param1, param2);
            break;
        default:
//            TAP_Print("\nEVT_???, event=%x, param1=%x, param2=%x", event, param1, param2);
            break;
    }
    return param1;
}

int main_menu_handler(type_menu *menu, dword param1)
{
    int selection;
    int i;
    int num_events;

    switch (param1) {
        case RKEY_Ok:
            selection = TAP_Win_GetSelection(&menu->win);
            DeleteMenu(menu);
            switch (selection) {
                case 0:
                    if (idle_state == STATE_IDLE) {
                        sprintf(strLog, "Loading epg.dat");
                        log_message(LOG_SCREEN_CONT|LOG_SERIAL, strLog, DELAY_SHORT);
                        //delete the epg
                        exTAP_EPG_DeleteAllEvents();
                        //load the data
                        load_epg("epg.dat");
                    }
                    else {
                        sprintf(strLog, "WARNING: busy loading data");
                        log_message(LOG_SCREEN_CONT|LOG_SERIAL, strLog, DELAY_MEDIUM);
                    }
                    return 0;
                case 1:
                    sprintf(strLog, "Saving epg.dat");
                    log_message(LOG_SCREEN_CONT|LOG_SERIAL, strLog, DELAY_SHORT);
                    save_epg("epg.dat");
                    return 0;
                case 2:
                    sprintf(strLog, "Saving MyStuff.mei");
                    log_message(LOG_SCREEN_CONT|LOG_SERIAL, strLog, DELAY_SHORT);
                    export_mei("MyStuff.mei");
                    return 0;
                case 3:
                    if (dsmcc_enabled) {
                        sprintf(strLog, "Removing DSMCC Hook");
                        log_message(LOG_SCREEN_CONT|LOG_SERIAL, strLog, DELAY_SHORT);
                        remove_dsmcc_hook();
                    }
                    else {
                        sprintf(strLog, "Applying DSMCC Hook");
                        log_message(LOG_SCREEN_CONT|LOG_SERIAL, strLog, DELAY_SHORT);
                        apply_dsmcc_hook();
                    }
                    return 0;

                case 4:
                    num_events = exTAP_EPG_GetNumEvents();
                    sprintf(strLog, "Number of Events = %d", num_events);
                    log_message(LOG_SCREEN_CONT|LOG_SERIAL, strLog, DELAY_LONG);
                    return 0;
                case 5:
                    close();
                    TAP_Exit();
                    return 0;
                case 6:
                    ShowMenu(&more_menu);
                    return 0;
            }
            break;
       case RKEY_Recall:
            DeleteMenu(menu);
            return param1; // cascades key to other TAP's
#ifdef KEY_REMAPPING
        case exRKEY_Exit:
#else
        case RKEY_Exit:
#endif
            DeleteMenu(menu);
            return 0; // exit without Cascading
        default:
            TAP_Win_Action(&menu->win, param1); // send all other key's to menu-Win
            return 0;
    }
    return param1;
}

int more_menu_handler(type_menu *menu, dword param1)
{
    int selection;
    int i;
    int num_events;
    dword heapSize, freeHeapSize, availHeapSize;

    switch (param1) {
        case RKEY_Ok:
            selection = TAP_Win_GetSelection(&menu->win);
            DeleteMenu(menu);
            switch (selection) {
                case 0:
                    if (eit_enabled) {
                        sprintf(strLog, "Removing EIT Hook");
                        log_message(LOG_SCREEN_CONT|LOG_SERIAL, strLog, DELAY_SHORT);
                        remove_eit_hook();
                    }
                    else {
                        sprintf(strLog, "Applying EIT Hook");
                        log_message(LOG_SCREEN_CONT|LOG_SERIAL, strLog, DELAY_SHORT);
                        apply_eit_hook();
                    }
                    return 0;
                case 1:
                    sprintf(strLog, "Deleting all Events");
                    log_message(LOG_SCREEN_CONT|LOG_SERIAL, strLog, DELAY_SHORT);
                    num_events = exTAP_EPG_DeleteAllEvents();
                    sprintf(strLog, " = %d", num_events);
                    log_message(LOG_SERIAL, strLog, DELAY_SHORT);
                    return 0;
                case 2:
                    if (epg_enabled) {
                        sprintf(strLog, "Disabling EPG data");
                        log_message(LOG_SCREEN_CONT|LOG_SERIAL, strLog, DELAY_SHORT);
                        TAP_ControlEit(0);
                        epg_enabled = 0;
                        strcpy(more_menu.item[3].str, "Enable EPG");
                    }
                    else {
                        sprintf(strLog, "Enabling EPG data");
                        log_message(LOG_SCREEN_CONT|LOG_SERIAL, strLog, DELAY_SHORT);
                        TAP_ControlEit(1);
                        epg_enabled = 1;
                        strcpy(more_menu.item[3].str, "Disable EPG");
                    }
                    return 0;
                case 3:
                    TAP_MemInfo(&heapSize, &freeHeapSize, &availHeapSize);
                    sprintf(strLog,"heapSize=%d    freeHeapSize=%d    availHeapSize=%d", heapSize, freeHeapSize, availHeapSize);
                    log_message(LOG_SCREEN_CONT|LOG_SERIAL, strLog, DELAY_LONG);
                    return 0;
                case 4:
                    sprintf(strLog,"Shutting Down");
                    log_message(LOG_SCREEN_CONT|LOG_SERIAL, strLog, DELAY_SHORT);
                    exTAP_Shutdown();
                    return 0;
                case 5:
                    {
                        word mjd;
                        byte hour, min, sec;

                        word mjd1;
                        byte hour1, min1;
                        dword utc;
                        dword local;

                        exTAP_GetUTCTime(&mjd1, &hour1, &min1);
                        TAP_GetTime(&mjd, &hour, &min, &sec);
                        utc = (mjd1<<16) + (hour1<<8) + min1;
                        local = exTAP_UTCtoLocalTime(utc);
                        sprintf(strLog, "UTC:      mjd=%04x, %s, hour=%02d:%02d    \nLOCAL:    mjd=%04x, %s, %02d:%02d:%02d\nUTC2LOCAL: mjd=%04x, %s, hour=%02d:%02d    ",
                                        mjd1,  mjd_to_string(mjd1), hour1, min1,
                                        mjd,  mjd_to_string(mjd), hour, min, sec,
                                        local>>16,  mjd_to_string(local>>16), (local>>8)&0xff, local&0xff);
                        log_message(LOG_SCREEN_CONT|LOG_SERIAL, strLog, DELAY_MEDIUM);
                    }
//                    dump_channels();
                    return 0;
            }
            break;
        case RKEY_Recall:
            DeleteMenu(menu);
            return param1; // cascades key to other TAP's
#ifdef KEY_REMAPPING
        case exRKEY_Exit:
#else
        case RKEY_Exit:
#endif
            DeleteMenu(menu);
            return 0; // exit without Cascading
        default:
            TAP_Win_Action(&menu->win, param1); // send all other key's to menu-Win
            return 0;
    }
    return param1;
}

int key_handler(dword param1)
{
    static int seen_shutdown = 0;

    switch(param1) {
        case RKEY_Power:
//            TAP_Print("\nRKEY_Power");
///*
//only if master shutdown tap
            if (shutdown == 0) {
                //RKEY_Power event manually generated
            }
            else {
                if (!seen_shutdown) {
                    sprintf(strLog,"Saving epg.dat");
                    log_message(LOG_SCREEN_CONT|LOG_SERIAL, strLog, DELAY_SHORT);
                    save_epg("epg.dat");
                    TAP_GenerateEvent(EVT_KEY, RKEY_Power, 0);
                    seen_shutdown = 1;
                } else {
                    sprintf(strLog,"Shutting Down");
                    log_message(LOG_SCREEN_WAIT|LOG_SERIAL, strLog, DELAY_SHORT);
                    exTAP_Shutdown();
                }
                shutdown++;
            }
//*/
/*
//if slave shutdown tap
            sprintf(strLog,"Saving epg.dat");
            log_message(LOG_SCREEN_CONT|LOG_SERIAL, strLog, DELAY_SHORT);
            save_epg("epg.dat");
*/
            break;

        //only if tsrcommander is not running
        case RKEY_Recall:
            if (!is_tsr_commander_running()) {
                ShowMenu(&main_menu);
                return 0;
            }
            break;
    }
    return param1;
}

//----------------------------------------------------------------------------
typedef struct
{
    byte signature[32];					//TFAMIPS
    dword id;
    byte padding1[28];
    const char name[MAX_PROGRAM_NAME];
    const char authorName[MAX_AUTHOR_NAME];
    const char description[MAX_DESCRIPTION];
    const char etcStr[MAX_ETCINFO_STR];
} type_TAPHeader;

typedef struct
{
    // From libtap.a
    dword _initData;
    void *TAP_Main;
    dword _tap_gotStart;
    dword _tap_gotEnd;
    dword _tap_textEnd;
    dword _tap_fixupStart;
    dword _tap_fixupEnd;
    void *TAP_EventHandler;
    dword _gp;
    dword _end;

    void *pTAP_SystemProc;
    void *_appl_version;
    void *spare0;
    void *pTAP_GetState;
    void *pTAP_ExitNormal;
    void *pTAP_EnterNormal;
    void *spare1;
    void *spare2;
    void *pTAP_Exit;
    void *pTAP_KbHit;
    void *pTAP_GetCh;
    void *pTAP_PutCh;
    void *pTAP_Print;
    void *pTAP_GetSystemProc;
    void *pTAP_undoc11;
    void *pTAP_SetBaudRate;
    //etc
} type_TAPEntryPoints;

typedef struct
{
    dword firmwaregp;
    dword tapgp;
    dword unknown1;
    type_TAPHeader *header;
    type_TAPEntryPoints *entryPoints;
    dword loadAddress;
    dword unknown2;
    dword zero;
} type_TAPProcess;

int is_tsr_commander_running(void)
{
    int i;
	type_TAPProcess *TAPProcess;

    for (i=0; i<16; i++) {
        TAPProcess = (type_TAPProcess *)(fw_tap_table + (i*32));
        if (TAPProcess->firmwaregp) {
            if (TAPProcess->header) {
                if (TAPProcess->header->id == 0x807a10b2) {  //tsr_commander
                    return 1;
                }
            }
        }
    }
    return 0;
}
//----------------------------------------------------------------------------
//menu stuff
dword message_showing = 0;
dword message_rgn = 0;

dword ShowMessage(char *lpMessage, char *lpMessage1, dword dwDelay, int wait)
{
	dword rgn;															// stores rgn-handle for osd
	dword w1,w2;														// stores width of message-text
	int i;
	int w = 430;
	char Message2[0x40];

	rgn = TAP_Osd_Create(0, 0, 720, 576, 0, FALSE );					// create rgn-handle
	w1 = TAP_Osd_GetW(lpMessage, 0, FNT_Size_1622 ) + 10;				// calculate width of message
	if (w1 > w) w1 = w;												    // if message is too long
	w2 = TAP_Osd_GetW(lpMessage1, 0, FNT_Size_1622 ) + 10;              // calculate width of message

	TAP_Osd_FillBox(rgn, (720-w)/2-5, 265, w+10, 54+22*(strlen(lpMessage1)/50), RGB(19,19,19) );	// draw background-box for border

	TAP_Osd_PutS(rgn, (720-w)/2, 270, (720+w)/2, lpMessage,
    	COLOR_White, RGB(3,5,10), 0, FNT_Size_1622, FALSE, ALIGN_CENTER);	 // show 1. message

    for(i=0; i<strlen(lpMessage1); i+=50) {
        memcpy(Message2, &lpMessage1[i], 50);
        Message2[50] = '\0';
    	TAP_Osd_PutS(rgn, (720-w)/2, 270+22*((i+50)/50), (720+w)/2, Message2,
	    	COLOR_White, RGB(3,5,10), 0, FNT_Size_1622, FALSE, ALIGN_CENTER);	 // show 2. message
	}
    if (wait) {
        message_showing = 0;
        TAP_Delay(dwDelay);                                             // show it for dwDelay /100 seconds
        TAP_Osd_Delete(rgn);                                            // release rgn-handle
        message_rgn = 0;
    }
    else {
        message_showing = TAP_GetTick() + dwDelay;
        message_rgn = rgn;
        return rgn;
    }
    return 0;
}

void ShowMenu(type_menu *menu)
{
    int i;

    if (!menu->visible) {
        menu->rgn = TAP_Osd_Create(0, 0, 720, 576, 0, FALSE);
        TAP_Win_SetDefaultColor(&menu->win);
        TAP_Win_Create(&menu->win, menu->rgn, 250, 210, 200, 80, FALSE, FALSE);
        TAP_Win_SetTitle(&menu->win, ptrfix(menu->fixme_titlep), 0, FNT_Size_1622);
//        TAP_Win_SetTitle(&menu->win, menu->title, 0, FNT_Size_1622);
        for (i=1; i<=menu->item[0].id; i++) {
            TAP_Win_AddItem(&menu->win, menu->item[i].str);
        }
        TAP_Win_SetSelection(&menu->win, 0);
        menu->visible = 1;
    }
}

void DeleteMenu(type_menu *menu)
{
    if (menu->visible) {
        TAP_Win_Delete(&menu->win);
        TAP_Osd_Delete(menu->rgn);
        menu->visible = 0;
    }
}
//----------------------------------------------------------------------------
//
//utils
//
void dump(dword addr, int length)
{
    int i,j;
    byte c;

    for (i=0; i<length; i+=0x10) {
        TAP_Print("\n%08x : ",addr+i);
        for (j=0; j<0x10; j++) {
            if ((i+j)<length) {
                TAP_Print("%02x ",*(byte *)(addr+i+j));
            }
            else {
                TAP_Print("   ");
            }
        }
        TAP_Print("  ");
        for (j=0; j<0x10; j++) {
            c = *(byte *)(addr+i+j);
            if ((i+j)<length) {
                TAP_Print("%c",((c>0x20) && (c<0x80))?c:' ');
            }
            else {
                TAP_Print(" ");
            }
        }
    }
    TAP_Print("\n");
}

int dump_channels(void)
{
    int i;
    int nTvSvc;
    int nRadioSvc;
    TYPE_TapChInfo channel_info;


    TAP_Channel_GetTotalNum(&nTvSvc, &nRadioSvc);
    for (i=0; i<nTvSvc; i++) {
        TAP_Channel_GetInfo(SVC_TYPE_Tv, i, &channel_info);
        TAP_Print("\n%2d, (%3d), %16s", i, channel_info.logicalChNum, channel_info.chName);
        TAP_Print(", onid:%04x, tsid:%04x, svcid:%04x", channel_info.orgNetId, channel_info.tsId, channel_info.svcId);
    }

    TAP_Channel_GetTotalNum(&nTvSvc, &nRadioSvc);
    for (i=0; i<nRadioSvc; i++) {
        TAP_Channel_GetInfo(SVC_TYPE_Radio, i, &channel_info);
        TAP_Print("\n%2d, (%3d), %16s", i, channel_info.logicalChNum, channel_info.chName);
        TAP_Print(", onid:%04x, tsid:%04x, svcid:%04x", channel_info.orgNetId, channel_info.tsId, channel_info.svcId);
    }

    return 0;
}

//still a bit crap
int get_channel_number(int service_id)
{
    int i;
    int nTvSvc, nRadioSvc;
    int num_channels;
    TYPE_TapChInfo chInfo;

    TAP_Channel_GetTotalNum(&nTvSvc, &nRadioSvc);
    for (i=0; i<nTvSvc; i++) {
        TAP_Channel_GetInfo(SVC_TYPE_Tv, i, &chInfo);
        if (service_id == chInfo.svcId) {
            return chInfo.logicalChNum;
        }
    }
    for (i=0; i<nRadioSvc; i++) {
        TAP_Channel_GetInfo(SVC_TYPE_Radio, i, &chInfo);
        if (service_id == chInfo.svcId) {
            return chInfo.logicalChNum;
        }
    }
    sprintf(strLog,"ERROR: get_channel_number() failed,  num_channels=%d, service_id=%04x", num_channels, service_id);
    log_message(LOG_SCREEN_CONT|LOG_SERIAL, strLog, DELAY_LONG);
    return -1;
}

//----------------------------------------------------------------------------
int init(void)
{
    int i,j;
    int ret;
    int channel_type, channel_number;
    TYPE_TapChInfo channel_info;
    word mjd;
    byte hour, min, sec;
    dword heapSize, freeHeapSize, availHeapSize;


    idle_state = STATE_IDLE;
    dsmcc_enabled = 0;
    eit_enabled = 0;
    epg_enabled = 1;

    main_menu.visible = 0;
    more_menu.visible = 0;

    TAP_MemInfo(&heapSize, &freeHeapSize, &availHeapSize);
    sprintf(strLog,"heapSize=%d    freeHeapSize=%d    availHeapSize=%d", heapSize, freeHeapSize, availHeapSize);
    log_message(LOG_SERIAL, strLog, DELAY_SHORT);

    init_crc();

#ifdef KEY_REMAPPING
    exTAP_Key_Extend(1);
#endif

    ret = apply_shutdown_hook();

    //increase the epg buffer size
    //short event from 0x80000->0x180000    (512K --> 1.5M)
    //extended event from 0x80000->0x100000    (512K --> 1M)
    exTAP_EPG_Resize(0x180000, 0x100000);

    //apply the hook before loading epg.dat
    sprintf(strLog,"Applying EIT Hook");
    log_message(LOG_SERIAL, strLog, DELAY_SHORT);
    apply_eit_hook();

    //load the epg
    if (idle_state == STATE_IDLE) {
        load_epg("epg.dat");
    }

    //enable dsmcc if tuned to 4tv and between 3-4am
    TAP_Channel_GetCurrent(&channel_type, &channel_number);
    TAP_Channel_GetInfo(channel_type, channel_number, &channel_info);
    //see if are tuned to 4tv
    if (channel_info.svcId == 0x6980) {
        //see if it is time to wake up
        TAP_GetTime(&mjd, &hour, &min, &sec);
        if (dsmcc_hook_id < 0) {
            if (hour==3) {
                sprintf(strLog,"Applying DSMCC Hook");
                log_message(LOG_SCREEN_CONT|LOG_SERIAL, strLog, DELAY_SHORT);
                apply_dsmcc_hook();
            }
        }
    }

    //recheck every 10 minutes
    ten_minutes = TAP_GetTick() + (10*60*100);

    TAP_MemInfo(&heapSize, &freeHeapSize, &availHeapSize);
    sprintf(strLog,"heapSize=%d    freeHeapSize=%d    availHeapSize=%d", heapSize, freeHeapSize, availHeapSize);
    log_message(LOG_SERIAL, strLog, DELAY_SHORT);
    return 0;
}

int close(void)
{
    dword heapSize, freeHeapSize, availHeapSize;

    sprintf(strLog,"Closing");
    log_message(LOG_SCREEN_WAIT|LOG_SERIAL, strLog, DELAY_SHORT);

    sprintf(strLog,"Saving epg.dat");
    log_message(LOG_SERIAL, strLog, DELAY_SHORT);
    save_epg("epg.dat");

    //all hooks MUST be removed before shutting down a tap
    sprintf(strLog,"Removing EIT Hook");
    log_message(LOG_SERIAL, strLog, DELAY_SHORT);
    remove_eit_hook();
    sprintf(strLog,"Removing DSMCC Hook");
    log_message(LOG_SERIAL, strLog, DELAY_SHORT);
    remove_dsmcc_hook();
    remove_shutdown_hook();

#ifdef KEY_REMAPPING
//    exTAP_Key_Extend(0);        //undo Key Mapping patch
#endif
    DeleteMenu(&more_menu);
    DeleteMenu(&main_menu);
    if (OpLog) TAP_Hdd_Fclose(OpLog);
    TSRCommanderErase();


    TAP_MemInfo(&heapSize, &freeHeapSize, &availHeapSize);
    sprintf(strLog,"heapSize=%d    freeHeapSize=%d    availHeapSize=%d", heapSize, freeHeapSize, availHeapSize);
    log_message(LOG_SERIAL, strLog, DELAY_SHORT);

    TAP_Print("\nClosing %s\n",__tap_program_name__);
    return 0;
}

//-----------------------------------------------------------------------------
//
//shutdown stuff
//
#define FW_SHUTDOWN_TASK_1209          0x80108c0c
#define FW_SHUTDOWN_IRQ_DISABLE_1209   0x80004da4
#define FW_SHUTDOWN_IRQ_ENABLE_1209    0x80004dd8

#define FW_SHUTDOWN_TASK_1225          0x80109098
#define FW_SHUTDOWN_IRQ_DISABLE_1225   0x80004de8
#define FW_SHUTDOWN_IRQ_ENABLE_1225    0x80004e1c

type_hook shutdown_hook_patch[] = {
//   firmware, enable, address,               length, original_data
    {0x1209,   1,      FW_SHUTDOWN_TASK_1209, 8,      {0x27, 0xbd, 0xff, 0xb8,  0xaf, 0xb4, 0x00, 0x38, }, },  //shutdown
    {0x1225,   1,      FW_SHUTDOWN_TASK_1225, 8,      {0x27, 0xbd, 0xff, 0xb8,  0xaf, 0xb4, 0x00, 0x38, }, },  //shutdown
};

int apply_shutdown_hook(void)
{
    switch (_appl_version) {
        case 0x1209:
            sprintf(strLog, "Applying Shutdown Hook");
            log_message(LOG_SERIAL, strLog, DELAY_SHORT);
            shutdown_hook_id = exTAP_ApplyHook(&shutdown_hook_patch[0]);
            break;
        case 0x1225:
            sprintf(strLog, "Applying Shutdown Hook");
            log_message(LOG_SERIAL, strLog, DELAY_SHORT);
            shutdown_hook_id = exTAP_ApplyHook(&shutdown_hook_patch[1]);
            break;
        default:
            sprintf(strLog, "ERROR: apply_shutdown_hook(), firmware version %04x not supported", _appl_version);
            log_message(LOG_SCREEN_WAIT|LOG_SERIAL, strLog, DELAY_LONG);
            return -1;
            break;
    }
    return 0;
}

int remove_shutdown_hook(void)
{
    int ret;

    if (shutdown_hook_id >= 0) {
        sprintf(strLog,"Removing Shutdown Hook");
        log_message(LOG_SERIAL, strLog, DELAY_SHORT);
        ret = exTAP_RemoveHook(shutdown_hook_id);
        if (ret == 0) shutdown_hook_id = -1;
    }
    return 0;
}

dword shutdown_hook(dword param1, dword param2)
{
    if (shutdown == 0) {
        //check firmware version
        switch (_appl_version) {
            case 0x1209:
                exTAP_CallFirmware((void *)FW_SHUTDOWN_IRQ_DISABLE_1209, 0, 0,0,0);       //disable shutdown irq, prevents automatic shutdown on timeout
            case 0x1225:
                exTAP_CallFirmware((void *)FW_SHUTDOWN_IRQ_DISABLE_1225, 0, 0,0,0);       //disable shutdown irq, prevents automatic shutdown on timeout
                break;
            default:
                sprintf(strLog, "ERROR: shutdown_hook(), firmware version %04x not supported", _appl_version);
                log_message(LOG_SCREEN_WAIT|LOG_SERIAL, strLog, DELAY_LONG);
                return param1;
                break;
        }
        shutdown ++;
        TAP_GenerateEvent(EVT_KEY, RKEY_Power, 0);
        return -1;                  //force rts; will not shutdown
    }
    //need this to ensure the RKEY_Power completes ...
    else
    if (shutdown == 1) {
        shutdown ++;
        return -1;                  //force rts; will not shutdown
    }
    else {
        shutdown = 0;
        return param1;              //allow continue; will shutdown
    }
    return param1;
}

//-----------------------------------------------------------------------------
//
//dmscc stuff
//
//hook the process_dsmcc() function
//ignore if table_id is not dsmcc (3b/3c)
//ignore if table has already been seen
//save the raw dsmcc section
//
type_hook dsmcc_hook_patch[] = {
//   firmware, enable, address,    length, original_data
    {0x1209,   1,      0x801aacfc, 8,      {0x27, 0xbd, 0xff, 0xe0,  0xaf, 0xb6, 0x00, 0x18, }, },  //process_dsmcc
    {0x1225,   1,      0x801ab730, 8,      {0x27, 0xbd, 0xff, 0x78,  0xaf, 0xb6, 0x00, 0x80, }, },  //process_dsmcc
};

int apply_dsmcc_hook(void)
{
    TYPE_TapChInfo  channel_info;
    int channel_type, channel_number;
    word mjd, year;
    byte month, day, weekDay;
    byte hour, min, sec;
    char filename[0x20] = "4tv.dsmcc";
    int ret;

    switch (_appl_version) {
        case 0x1209:
            dsmcc_hook_id = exTAP_ApplyHook(&dsmcc_hook_patch[0]);
            break;
        case 0x1225:
            dsmcc_hook_id = exTAP_ApplyHook(&dsmcc_hook_patch[1]);
            break;
        default:
            sprintf(strLog, "ERROR: apply_dsmcc_hook(), firmware version %04x not supported", _appl_version);
            log_message(LOG_SCREEN_WAIT|LOG_SERIAL, strLog, DELAY_LONG);
            return -1;
            break;
    }

    num_tables = 0;
    table = (type_table *)malloc(sizeof(type_table) * MAX_TABLE_ID);
    if (table == 0) {
        sprintf(strLog, "ERROR: apply_dsmcc_hook(), malloc failed, table=%x, len=%x", table, sizeof(type_table) * MAX_TABLE_ID);
        log_message(LOG_SCREEN_WAIT|LOG_SERIAL, strLog, DELAY_LONG);
    }
    memset(table, 0, sizeof(type_table) * MAX_TABLE_ID);


    if (!dsmcc_Op) {
        //create the filename based on channel_name, start_time
        TAP_Channel_GetCurrent(&channel_type, &channel_number);
        TAP_Channel_GetInfo(channel_type, channel_number, &channel_info);
        TAP_GetTime(&mjd, &hour, &min, &sec);
        TAP_ExtractMjd(mjd, &year, &month, &day, &weekDay);
        sprintf(filename, "%s_%04d%02d%02d%02d%02d.dsmcc", channel_info.chName, year, month, day, hour, min);
        TAP_Hdd_ChangeDir("/");
        TAP_Hdd_ChangeDir("ProgramFiles");
        //if exists delete it
        if (TAP_Hdd_Exist(filename)) {
            TAP_Hdd_Delete(filename);
        }
        ret = TAP_Hdd_Create(filename, ATTR_NORMAL);
        dsmcc_Op = TAP_Hdd_Fopen(filename);
    }
    dsmcc_enabled = 1;
    strcpy(main_menu.item[4].str, "Remove DSMCC Hook");

    return 0;
}

int remove_dsmcc_hook(void)
{
    int ret;

TAP_Print("\nnum_tables=%x",num_tables);
    if (dsmcc_hook_id >=0) {
        ret = exTAP_RemoveHook(dsmcc_hook_id);
        if (ret == 0) dsmcc_hook_id = -1;
    }
    if (dsmcc_Op) {
        TAP_Hdd_Fclose(dsmcc_Op);
        dsmcc_Op = 0;
    }

    free(table);
    dsmcc_enabled = 0;
    strcpy(main_menu.item[4].str, "Apply DSMCC Hook");
    return 0;
}

dword dsmcc_hook(dword param1, dword param2)
{
    type_hook_registers *reg;
    byte *buf;
    int section_length;

//TAP_Print("\nEVT_HOOK_dsmcc, param1=%x, param2=%x", param1, param2);
    reg = (type_hook_registers *)param2;
    if ((reg->a0 > 0x80000000) && (reg->a0 < 0x84000000)) {
        buf = (byte *)(reg->a0);
        section_length = 3 + (((buf[1]<<8) + buf[2]) & 0xfff);


        section_process_3b3c(buf, section_length);
    }
    return param1;
}

//for dsmcc
int section_process_3b3c(byte *buf, int len)
{
    int i;
    int ret;
    byte  table_id;
    word section_length;
    word table_id_extension;
    byte version_number;
    byte current_next_indicator;
    byte section_number;
    byte last_section_number;

    int found;
    type_table *t;

    table_id               = buf[0];
    section_length         = 3 + (((buf[1]<<8) + buf[2]) & 0xfff);
    table_id_extension     = ((buf[3]<<8) + buf[4]) & 0xffff;
    version_number         = (buf[5]>>1) & 0x1f;
    current_next_indicator = (buf[5]>>0) & 0x1;
    section_number         = buf[6];
    last_section_number    = buf[7];

    //ignore if not dsmcc
    if ((table_id != 0x3b) && (table_id != 0x3c)) {
        sprintf(strLog,"WARNING: section_process_3b3c(), invalid table_id = %02x", table_id);
        log_message(LOG_SCREEN_CONT|LOG_SERIAL, strLog, DELAY_SHORT);
        return 0;
    }

    //there may be multiple sections
    //each table has the same table_id, table_id_extension
    //use last_section_number to determine the number of sections per table
    //process each section independantly
    //but keep a count

    found = -1;
    for (i=0; i<MAX_TABLE_ID; i++) {
        if ((table[i].table_id == table_id) && (table[i].table_id_extension == table_id_extension) && (table[i].section_number == section_number)) {
            t = (type_table *)&table[i];
//            TAP_Print("\nDSMCC-%02x:%04x %3d/%3d: matched table, %x", table_id, table_id_extension, section_number, last_section_number, i);
            TAP_Print("m");
            found = i;
            goto cont1;
        }
    }
cont1:
    if (found < 0) {
        t = (type_table *)&table[num_tables];
//        TAP_Print("\nDSMCC-%02x:%04x %3d/%3d: newtable, %x", table_id, table_id_extension, section_number, last_section_number, num_tables);
        TAP_Print("n");
        if (num_tables >= MAX_TABLE_ID) {
            sprintf(strLog,"ERROR: DSMCC-%02x:%04x %3d/%3d: out of tables - %x", num_tables, table_id, table_id_extension, section_number, last_section_number);
            log_message(LOG_SCREEN_WAIT|LOG_SERIAL, strLog, DELAY_LONG);
            return 0;
        }
        num_tables++;
    }

    t->table_id = table_id;
    t->table_id_extension = table_id_extension;
    t->section_number = section_number;
    t->last_section_number = last_section_number;
    t->section_length = section_length;
    t->len = len;
    t->count ++;


    //have we already processed this section
    //has it changed
    if (t->valid) {
        if (t->version_number == version_number) {
//            TAP_Print("\nDSMCC-%02x:%04x %3d/%3d: already got section", table_id, table_id_extension, section_number, last_section_number);
            return 0;
        }
        else {
            TAP_Print("\nDSMCC-%02x:%04x %3d/%3d: new version, was %2x, now %2x", table_id, table_id_extension, section_number, last_section_number, t->version_number, version_number);
            t->valid = 0;
        }
    }

    t->valid = 1;
    t->version_number = version_number;

    //save everything
    if (dsmcc_Op) {
        TAP_Print("s");
        TAP_Hdd_Fwrite((byte *)&buf[0], len, 1, dsmcc_Op);
    }

    return 0;
}

//-----------------------------------------------------------------------------
//
//eit stuff
//
//hook the process_eit() function
//this code just filters by service_id
//allow through all now/next (table_id=0x4e)
//only allow through enabled schedule (table_id=50/51)
//could add max# days per channels
//should never allow if servicve_id does not exists - e.g when 4tv data
//
type_hook eit_hook_patch[] = {
//   firmware, enable, address,    length, original_data
    {0x1209,   1,      0x8014ed60, 8,      {0x27, 0xbd, 0xff, 0x28,  0xaf, 0xb0, 0x00, 0xb8, }, },  //process_eit
    {0x1225,   1,      0x8014fc2c, 8,      {0x27, 0xbd, 0xff, 0x28,  0xaf, 0xb0, 0x00, 0xb8, }, },  //process_eit
};

byte *to;

int apply_eit_hook(void)
{
    int i;

    switch (_appl_version) {
        case 0x1209:
            eit_hook_id = exTAP_ApplyHook(&eit_hook_patch[0]);
            break;
        case 0x1225:
            eit_hook_id = exTAP_ApplyHook(&eit_hook_patch[1]);
            break;
        default:
            sprintf(strLog, "ERROR: apply_eit_hook(), firmware version %04x not supported", _appl_version);
            log_message(LOG_SCREEN_WAIT|LOG_SERIAL, strLog, DELAY_LONG);
            return -1;
            break;
    }

    to = (byte *)malloc(0x1000);
    if (to == 0) {
        sprintf(strLog, "ERROR: apply_eit_hook(), malloc failed, to = %x, len=%x", to, 0x1000);
        log_message(LOG_SCREEN_WAIT|LOG_SERIAL, strLog, DELAY_LONG);
    }

    eit_enabled = 1;
    strcpy(more_menu.item[1].str, "Remove EIT Hook");
    return 0;
}

int remove_eit_hook(void)
{
    int ret;

    if (eit_hook_id >=0) {
        ret = exTAP_RemoveHook(eit_hook_id);
        if (ret == 0) {
            eit_hook_id = -1;
        }
    }
    free(to);

    eit_enabled = 0;
    strcpy(more_menu.item[1].str, "Apply EIT Hook");
    return 0;
}

dword eit_hook(dword param1, dword param2)
{
    type_hook_registers *reg;

//TAP_Print("\nEVT_HOOK_eit, param1=%x, param2=%x", param1, param2);
    reg = (type_hook_registers *)param2;
    if ((reg->a0 > 0x80000000) && (reg->a0 < 0x84000000)) {
        reg->a0 = service_id_filter((byte *)(reg->a0));
    }
    return param1;
}


typedef struct {
    int enabled;
    word service_id;
}
type_service_filter;

type_service_filter service_filter[] = {
    1,  0x1044,         //        BBC ONE
    1,  0x1084,         //        BBC TWO
    1,  0x2045,         //           ITV1
    1,  0x20c0,         //      Channel 4
    1,  0x3241,         //           five
    1,  0x3242,         //           five
    1,  0x2085,         //           ITV2
    1,  0x10ff,         //      BBC THREE
    1,  0x2144,         //              8
    1,  0x2145,         //       Teletext
    1,  0x41c0,         //       BBC FOUR
    1,  0x56c0,         //     Sky Travel
    1,  0x5700,         //   UKTV History
    1,  0x20fa,         //         More 4
    1,  0x2100,         //             E4
    1,  0x3780,         //           abc1
    1,  0x3340,         //            QVC
    1,  0x3cc0,         //      UKTV Gold
    1,  0x6440,         //       The HITS
    1,  0x64c0,         // UKTV Br'tIdeas
    1,  0x6500,         //           f tn
    1,  0x6480,         //            TMF
    1,  0x6540,         //    Ideal World
    1,  0x37c0,         //         bid tv
    1,  0x3d00,         //  price-drop tv
    1,  0x3a00,         //            TCM
    1,  0x3a80,         //     UKTV Style
    1,  0x3b40,         //      Discovery
    1,  0x3b00,         // DiscoveryRTime
    1,  0x3e40,         //      UKTV Food
    1,  0x123f,         //   CBBC Channel
    1,  0x20a1,         //           ITV4
    1,  0x4240,         //       CBeebies
    1,  0x3ac0,         //    Cartoon Nwk
    1,  0x3c80,         //      Boomerang
    1,  0x206d,         //           ITV3
    1,  0x2066,         //           ITV3
    1,  0x3d80,         //       Xtraview
    1,  0x69c0,         //   Men & Motors
    1,  0x5740,         //           E4+1
    1,  0x113f,         //    BBC NEWS 24
    1,  0x2185,         //       ITV News
    1,  0x5640,         //       Sky News
    1,  0x5680,         //  Sky Spts News
    1,  0x3e00,         //      Bloomberg
    1,  0x4180,         //    BBC PARLMNT
    1,  0x4e00,         //      Community
    1,  0x3dc0,         //   Teachers' TV
    1,  0x3e80,         //   Eurosport UK
    1,  0x2240,         //  Teletext on 4
    1,  0x117f,         //           BBCi
    1,  0x6940,         //  YooPlay Games
    1,  0x39c0,         // Ttext Holidays
    1,  0x21c3,         //  Teletext Cars
    1,  0x3ec0,         //        Toonami
    1,  0x3b80,         //   Television X
    1,  0x3d40,         //        Red Hot
    1,  0x5c40,         //  YooPlay Games
    1,  0x4c00,         //            301
    1,  0x4c40,         //            302
    1,  0x4c80,         //            303
    1,  0x6980,         // 4TVinteractive
    1,  0x2104,         //      Quiz Call
    1,  0x3800,         //    BBC Radio 1
    1,  0x4700,         //      1Xtra BBC
    1,  0x3840,         //    BBC Radio 2
    1,  0x3880,         //    BBC Radio 3
    1,  0x38c0,         //    BBC Radio 4
    1,  0x4600,         //    BBC R5 Live
    1,  0x4640,         // BBC 5L SportsX
    1,  0x4680,         //    BBC 6 Music
    1,  0x46c0,         //          BBC 7
    1,  0x4740,         // BBC Asian Net.
    1,  0x6780,         //  BBC World Sv.
    1,  0x67c0,         // The Hits Radio
    1,  0x6680,         //    Smash Hits!
    1,  0x6640,         //           Kiss
    1,  0x3900,         //           heat
    1,  0x6800,         //          Magic
    1,  0x6840,         //              Q
    1,  0x6740,         //        oneword
    1,  0x6700,         //      smooth fm
    1,  0x3940,         //           MOJO
    1,  0x66c0,         //       Kerrang!
    1,  0x5840,         //      talkSPORT
    1,  0x5a00,         //             3C
    1,  0x59c0,         //  Premier Radio
} ;


/*
int init_service_id_filter()
{
    int i;
    int num_service_filters;

    num_service_filters = sizeof(service_filter) / sizeof(type_service_filter);
    service_hash = open_hash(2, num_service_filters);
    for (i=0; i<num_service_filters; i++) {
        if (service_filter[i].enabled) {
            addhash(service_hash, (byte *)&service_filter[i].service_id, i);
        }
    }
    return 0;
}

int close_service_id_filter()
{
    close_hash(service_hash);
    service_hash = 0;
    return 0;
}
*/

dword service_id_filter(byte *from)
{
    dword crc;
    byte table_id;
    word section_length;
    word table_id_extension;
    byte version_number;
    byte current_next_indicator;
    byte section_number;
    byte last_section_number;
    word transport_stream_id;
    word original_network_id;

    int i;
//    int num_service_filters;
//    int reject;
//    type_hash_element *hashptr;
    int to_len;

    //do crc check
    section_length         = 3 + (((from[1]<<8) + from[2]) & 0xfff);
    crc = crc32(from, section_length);
    if (crc) {
        TAP_Print("\nWARNING: *** CRC ERROR ***, section_length=%x", section_length);
        dump((dword)&from[0], 0x40);
        return (dword)from;
    }

    table_id               = from[0];
    section_length         = 3 + (((from[1]<<8) + from[2]) & 0xfff);
    table_id_extension     = ((from[3]<<8) + from[4]) & 0xffff;
    version_number         = (from[5]>>1) & 0x1f;
    current_next_indicator = (from[5]>>0) & 0x1;
    section_number         = from[6];
    last_section_number    = from[7];
    transport_stream_id    = ((from[8]<<8) + from[9]) & 0xffff;         //eit only
    original_network_id    = ((from[10]<<8) + from[11]) & 0xffff;       //eit only

//    TAP_Print("\nEIT:  0012:%02x:%04x %3d/%3d: %08x:%x", table_id, table_id_extension, section_number, last_section_number, buf, section_length);

/*
    //need to filter service_id, otherwise we will end up with lots of 'hidden' entries
    //match service id
    num_service_filters = sizeof(service_filter) / sizeof(type_service_filter);
//TAP_Print("\nnum_service_filters=%d", num_service_filters);
    reject = 1;
    for (i=0; i<num_service_filters; i++) {
        if (service_filter[i].enabled) {
            if (service_filter[i].service_id == table_id_extension) {
                reject = 0;
                break;
            }
        }
    }
    if (reject) {
//TAP_Print("\nrejecting: %02x:%04x", table_id, service_id);
        section_length = 0;
        from[1] = (section_length>>8) & 0xff;
        from[2] = (section_length>>0) & 0xff;
        return (dword)from;
    }
*/

    //reject unless eit
    switch (table_id) {
        case 0x4e:
        case 0x4f:
            //allow all now_next
            break;
        case 0x50:
        case 0x51:
        case 0x60:
        case 0x61:
            //only allow through service_id that match our filter
            //to zap, either set the length < 0x10, or break the crc
/*
            //match the service id
            hashptr = findhash(service_hash, (byte *)&table_id_extension);
            if (!hashptr) {
                //if fails service filter, trash the eit, will also trash the crc
//                TAP_Print("\n%04x-%04x-%04x %02x:%04x %3d/%3d: service filtered", 0x0012, original_network_id, transport_stream_id, table_id, table_id_extension, section_number, last_section_number);
                section_length = 0;
                from[1] = (section_length>>8) & 0xff;
                from[2] = (section_length>>0) & 0xff;
                return (dword)from;
            }
*/
            break;
        default:
            sprintf(strLog,"WARNING: service_id_filter(), invalid table_id = %02x", from[0]);
            log_message(LOG_SCREEN_CONT|LOG_SERIAL, strLog, DELAY_SHORT);
            return (dword)from;
            break;
    }

    //now process eit section
    convert_event_information_section(from, to, &to_len);

    if (to_len > 0x1000) {
        TAP_Print("\nERROR: service_id_filter(), to_len=%x",to_len);
//        dump((dword)&to[0], to_len);
        dump((dword)&to[0], 0x40);
    }

    return (dword)&to[0];
}

//----------------------------------------------------------------------------
//
//load the raw epg data from disk
//the TAP_PutEvent() function is incredibly slow, so use idle time
//
int load_epg(char *filename)
{
    static int i = 0;
    static int last_i = -1;
    static int num_events = 0;
    static type_eventtable *et = 0;
    static type_eventbuffer ebuf = {0};
    static byte *pkt;
    int plen;

    int short_blen;
    dword short_fbase;
    dword short_flen;
    static char *short_buffer = 0;

    int extended_blen;
    dword extended_fbase;
    dword extended_flen;
    static char *extended_buffer = 0;

    word mjd;
    byte hour, min, sec;
    dword now;
//v0.02
    dword start;
    dword end;

    int err;
    TYPE_File *Ip;
    int offset;
	type_TAPProcess *TAPProcess;

    switch (idle_state) {
        case STATE_IDLE:
            //get some memory for the input data
            et = (type_eventtable *)malloc(FW_MAX_EVENTS*0x40);
            if (et == 0) {
                sprintf(strLog, "ERROR: load_epg(), malloc failed, et=%x, len=%x", et, FW_MAX_EVENTS*0x40);
                log_message(LOG_SCREEN_WAIT|LOG_SERIAL, strLog, DELAY_LONG);
                idle_state = STATE_IDLE;
                break;
            }

            short_blen = exTAP_EPG_GetSize(0);
            short_buffer = (char *)malloc(short_blen);
            if (short_buffer == 0) {
                sprintf(strLog, "ERROR: load_epg(), malloc failed, short_buffer=%x, len=%x", short_buffer, short_blen);
                log_message(LOG_SCREEN_WAIT|LOG_SERIAL, strLog, DELAY_LONG);
                free(et);
                idle_state = STATE_IDLE;
                break;
            }

            extended_blen = exTAP_EPG_GetSize(1);
            extended_buffer = (char *)malloc(extended_blen);
            if (extended_buffer == 0) {
                sprintf(strLog, "ERROR: load_epg(), malloc failed, extended_buffer=%x, len=%x", extended_buffer, extended_blen);
                log_message(LOG_SCREEN_WAIT|LOG_SERIAL, strLog, DELAY_LONG);
                free(short_buffer);
                free(et);
                idle_state = STATE_IDLE;
                break;
            }

TAP_Print("\nShort Event Len = %x, Extended Event Len = %x", short_blen, extended_blen);

            //and some for the constructed eit
            pkt = (byte *)malloc(0x1000);
            if (pkt == 0) {
                sprintf(strLog, "ERROR: load_epg(), malloc failed, pkt=%x, len=%x", pkt, 0x1000);
                log_message(LOG_SCREEN_WAIT|LOG_SERIAL, strLog, DELAY_LONG);
               free(extended_buffer);
                free(short_buffer);
                free(et);
                idle_state = STATE_IDLE;
                break;
            }


            //load from disk
            sprintf(strLog, "Reading %s ...", filename);
            log_message(LOG_SERIAL, strLog, DELAY_SHORT);
            TAP_Hdd_ChangeDir("/");
            TAP_Hdd_ChangeDir("ProgramFiles");
            Ip = TAP_Hdd_Fopen(filename);
            if (!Ip) {
                sprintf(strLog, "ERROR: load_epg(), %s not found", filename);
                log_message(LOG_SERIAL, strLog, DELAY_SHORT);
                free(pkt);
                free(extended_buffer);
                free(short_buffer);
                free(et);
                idle_state = STATE_IDLE;
                return -1;
            }
            TAP_Hdd_Fread(et, FW_MAX_EVENTS*0x40, 1, Ip);
            TAP_Hdd_Fread(&ebuf, sizeof(type_eventbuffer), 1, Ip);                  //event buffer

            if (short_blen < 32*ebuf.short_event_blocks) {
                sprintf(strLog,"ERROR: load_epg(), short buffer too small, require %x, got %x", 32*ebuf.short_event_blocks, short_blen);
                log_message(LOG_SCREEN_WAIT|LOG_SERIAL, strLog, DELAY_LONG);
                TAP_Hdd_Fclose(Ip);
                free(pkt);
                free(extended_buffer);
                free(short_buffer);
                free(et);
                idle_state = STATE_IDLE;
                return -1;
            }
            else {
                TAP_Hdd_Fread(short_buffer, 32*ebuf.short_event_blocks, 1, Ip);
            }

            if (extended_blen < 32*ebuf.extended_event_blocks) {
                sprintf(strLog,"ERROR: load_epg(), extended buffer too small, require %x, got %x", 32*ebuf.extended_event_blocks, extended_blen);
                log_message(LOG_SCREEN_WAIT|LOG_SERIAL, strLog, DELAY_LONG);
                TAP_Hdd_Fclose(Ip);
                free(pkt);
                free(extended_buffer);
                free(short_buffer);
                free(et);
                idle_state = STATE_IDLE;
                return -1;
            }
            else {
                TAP_Hdd_Fread(extended_buffer, 32*ebuf.extended_event_blocks, 1, Ip);
            }

            TAP_Hdd_Fclose(Ip);

            //now add them to the internal EPG
            num_events = 0;
            i = 0;
            idle_state = STATE_LOAD_EPG_BUSY;
            break;
        case STATE_LOAD_EPG_BUSY:
            if (et[i].char00 & 0xc0) {
                //convert the pointers
                if (et[i].event_name) {
                    offset = (dword)et[i].event_name - (dword)ebuf.short_event_buffer;
                    et[i].event_name = &short_buffer[offset];

                    if (et[i].extended_event_name) {
                        offset = (dword)et[i].extended_event_name - (dword)ebuf.extended_event_buffer;
                        if (et[i].num_extended > 0) {
                            et[i].extended_event_name = &extended_buffer[offset];
                        }
                    }
                }

                //ignore if event already over
//                TAP_GetTime(&mjd, &hour, &min, &sec);
//we want this in UTC time ...
                exTAP_GetUTCTime(&mjd, &hour, &min);
                now = (mjd << 16) + (hour*60) + (min);
//v0.02
                start = (et[i].start_date << 16) + (et[i].start_hour*60) + (et[i].start_min);
                end = (et[i].end_date << 16) + (et[i].end_hour*60) + (et[i].end_min);
                if (now > end) {
                    //ignore it
                }

                //convert event to eit section
                else
                if (et[i].event_name) {
                    if ((num_events % 100) == 0) TAP_Print(".");

                    et[i].char00 &= 0xe0;     //clear running status
//v0.02
                    if (now >= start) {
                        et[i].char00 |= 0x4;    //set running status
                    }

//test-jam parental rating
//                    et[i].parental_rating = 12;

                    plen = set_event_information_section((type_eventtable *)&et[i], pkt);

                    //ideally should pack multiple events in each section

if (plen > 1000) {
TAP_Print("\npkt=%x, plen=%x", pkt, plen);
TAP_Print("\noh dear ...");
dump((dword)pkt, 0x100);
}
else {

                    err = TAP_PutEvent(pkt);    //this will return 0 if the section has a crc error etc
                    if (!err) {
                        TAP_Print("\nWARNING: TAP_PutEvent(), err=%x (0=fail)",err);
                    }
}
                    num_events ++;
                }
                else {
                    TAP_Print("\nERROR: et[%d].event_name=%x", i, et[i].event_name);
                }
            }
            i++;
            if (i >= FW_MAX_EVENTS) idle_state = STATE_LOAD_EPG_CLOSE;
            break;
        case STATE_LOAD_EPG_CLOSE:
            sprintf(strLog, "Loaded epg.dat.  Added %d events", num_events);
            log_message(LOG_SCREEN_CONT|LOG_SERIAL, strLog, DELAY_SHORT);

            //free the memory
            free(pkt);
            free(extended_buffer);
            free(short_buffer);
            free(et);
            idle_state = STATE_IDLE;
            break;
    }

    return 0;
}

//
//save the internal EPG to disk
//want this to be fast, since it will execute on a shutdown
//
int save_epg(char *filename)
{
    int num_events;
    TYPE_File *Op;

    num_events = exTAP_EPG_GetNumEvents();
TAP_Print(", %d events ...", num_events);
    TAP_Hdd_ChangeDir("/");
    TAP_Hdd_ChangeDir("ProgramFiles");
    TAP_Hdd_Create(filename, ATTR_NORMAL);
    Op = TAP_Hdd_Fopen(filename);
    if (!Op) {
        sprintf(strLog,"ERROR: save_epg(), cannot open %s", filename);
        log_message(LOG_SCREEN_WAIT|LOG_SERIAL, strLog, DELAY_LONG);
        return -1;
    }

    TAP_Hdd_Fwrite(&fw_event_table[0], FW_MAX_EVENTS*0x40, 1, Op);                                          //event table
    TAP_Hdd_Fwrite(fw_event_buffer, sizeof(type_eventbuffer), 1, Op);                                       //event buffer
    TAP_Hdd_Fwrite(fw_event_buffer->short_event_buffer, 32*fw_event_buffer->short_event_blocks, 1, Op);       //short event data
    TAP_Hdd_Fwrite(fw_event_buffer->extended_event_buffer, 32*fw_event_buffer->extended_event_blocks, 1, Op); //extended event data

    TAP_Hdd_Fclose(Op);
TAP_Print(" done");
    return 0;
}



//----------------------------------------------------------------------------
//mei file format (from rt2mei.php)
//
//delimited with |
//                                   e.g.
//      line_length                 (000234)
//      program start time          (200512220600)
//      program stop time           (200512220915)
//      LCN of channel              (appears to be rt channel id?)
//      program title               (text)
//      secondary title, e.g. episode title (text)
//      if this is a repeat         (<blank>,TRUE)
//      if this is first screening of a new show - not in RT data (<blank>,TRUE)
//      production date             (1972)
//      if it is a premiere         (<blank>,Premiere)
//      last chance                 (<blank>)
//      program description         (text)
//      program genre               (text)
//      program length              (minutes)
//      widescreen                  (<blank>,16:9)
//      whether it is in colour     (<blank>,no)
//      whether there are subtitles (<blank>,teletext)
//      episode number              (<blank>,3/3)
//      episodes start at 1         (<blank>,xmltv_ns)
//      rating system               (<blank>,BBFC)
//      rating                      (<blank>,U,PG,12,15)
//      star rating                 (<blank>,4/5)
//
//problems with .mei:
//  lack of documentation
//  lack of features
//  lack of extensibility
//      need a defined meachanism to add further fields _without_ breaking existing applications
//  length field - why? pointless ...
//  how to id channel
//      lcn/rtid field - which, what does it mean
//  lack of
//      original network id
//      transport stream id
//      service id
//      event id
//  end_time or duration; not both, what if they differ?
//  meaning of 'epsiode number', 'xmltv_ns' etc
//  inconsistancies ...
//      repeat - <blank, TRUE>
//      premier - <blank, Premiere>
//      widescreen - <blank, 16:9>
//  need a definitive list of possibilities for these boolean? fields
//
//
//
//buried deep on p15 of the MySuff_Extended_Info thread ...
//http://www.toppy.org.uk/forum/viewtopic.php?t=2115&postdays=0&postorder=asc&start=214
//
//MEI has the following info in each of it's columns (separated by a "|"):
//column - field - desc
//0 - xxxxxx - 6 char string indicating the total number of chars on the line (counting from field 1 onwards)
//1 - prog_start_time - program start time, a 12 digit string yyyymmddhhmm
//2 - prog_stop_time - program stop time, a 12 digit string yyyymmddhhmm
//3 - prog_channel - Freeview LCN
//4 - prog_title - program title, eg "The Simpsons"
//5 - prog_sub_title - program sub title, eg "Datalore"
//6 - prog_prev_shown - repeat?
//7 - prog_new - new program?
//8 - prog_date - Production date (is in there for films)
//9 - prog_premiere - Is it a film Premiere
//10 - prog_last_chance - don't think it's used by Radio Times (but identifies last chance)
//11 - prog_desc - COMPLETE Program Description (longest I found is over 1700 chars)
//12 - prog_category - Genre
//13 - prog_length - Length in minutes
//14 - prog_aspect - 16:9 or 4:3
//15 - prog_colour - Colour if blank, B&W (if B&W)
//16 - prog_subtitles - Subtitles type (onscreen if for foreign language film, or Teletext if "optional")
//17 - prog_episode_num - eg 1/6
//18 - prog_episode_system - either blank, onscsreen or mei_xmltv_ns**
//19 - prog_rating_system - ie BBFC for films
//20 - prog_rating - ie PG, 15, 18 etc
//21 - prog_star_rating - ie 2/5, 5/5 etc
//
//[edited 13/12/2005]
//**I decided to clean up the xmltv_ns episode numbers.
//If MEI has processed episode numbers that were in the xmltv_ns format, this field will say "mei_xmltv_ns". If no episode number is available it will be blank. MEI will not clean up episode numbers if they were provided in the "onscreen" format - in which case they will be contained in the prog_episode_num field, and prog_episode_system will indicate this as "onscreen".
//[/edit]
//


//
//extract genre from extended descriptor ...
//
int get_genre(type_eventtable *et, char **value)
{
    int i,j;

    if (et->extended_event_name == 0) return -1;
    if (et->num_extended == 0) return -1;
    j = 0;
    for (i=0; i<et->num_extended; i++) {
        if (strncmp(&et->extended_event_name[j], "genre", 5) == 0) {
            j += strlen(&et->extended_event_name[j]);        //skip name
            j ++;                                            //skip zero
            *value = &et->extended_event_name[j];
            j += strlen(&et->extended_event_name[j]);        //skip value
            j ++;                                            //skip zero
            return 0;
        }
        else {
            j += strlen(&et->extended_event_name[j]);        //skip name
            j ++;                                            //skip zero
            j += strlen(&et->extended_event_name[j]);        //skip value
            j ++;                                            //skip zero
        }
    }
    return -2;
}

//
//dump internal epg to mei file
//
int export_mei(char *filename)
{
    TYPE_File *Op;
    int num_events;
    type_eventtable *et;
    int mjd;
    byte weekDay, day, month;
    word year;
    word end_time;
    int i,j,k;
    int err;
    char *line;
    char length[0x10];
    int lcn;
    char *genre;
    dword utc;
    dword local;

    //create the mei file in ram, since fwrite is dog slow
    //but how long?
    line = malloc(0x100000);      //1 meg
    if (line == 0) {
        sprintf(strLog, "ERROR: export_mei(), malloc failed, line=%x, len=%x", line, 0x100000);
        log_message(LOG_SCREEN_WAIT|LOG_SERIAL, strLog, DELAY_LONG);
        return -1;
    }
    sprintf(strLog, "Saving %s ...", filename);
    log_message(LOG_SERIAL, strLog, DELAY_SHORT);
    TAP_Hdd_ChangeDir("/");
    TAP_Hdd_ChangeDir("ProgramFiles");
    //if exists delete it
    if (TAP_Hdd_Exist(filename)) {
        TAP_Hdd_Delete(filename);
    }
    TAP_Hdd_Create(filename, ATTR_NORMAL);
    Op = TAP_Hdd_Fopen(filename);
    if (!Op) {
        sprintf(strLog, "ERROR: export_mei(), cannot open file, %s", filename);
        log_message(LOG_SCREEN_WAIT|LOG_SERIAL, strLog, DELAY_LONG);
        free(line);
        return -1;
    }

    et = fw_event_table;
    num_events = 0;
    j = 0;
    for (i=0; i<FW_MAX_EVENTS; i++) {
        if (et[i].char00 & 0xc0) {
            if ((num_events % 100) == 0) TAP_Print(".");

            //flush to disk   - assumes max line length = 1024 ... !!!
//v0.02
//            if (j > (0x400000-0x400)) {
            if (j > (0x100000-0x400)) {
                TAP_Print("\ni=%d, saving %x bytes ...", i, j);
                TAP_Hdd_Fwrite(line, j, 1, Op);
                j = 0;
            }

            k = j;
            // line length
            j += 7;                     //length


            // program start time
//            TAP_ExtractMjd(et[i].start_date, &year, &month, &day, &weekDay);
//            sprintf(&line[j],"%04d%02d%02d%02d%02d|", year, month, day, et[i].start_hour, et[i].start_min);
//want to export in local time
//need to convert et[i].start_date/time to local time
            utc = (et[i].start_date << 16) + (et[i].start_hour<<8) + et[i].start_min;
            local = exTAP_UTCtoLocalTime(utc);
            TAP_ExtractMjd(local>>16, &year, &month, &day, &weekDay);
            sprintf(&line[j],"%04d%02d%02d%02d%02d|", year, month, day, (local>>8)&0xff, local&0xff);

            j += strlen(&line[j]);
            // program stop time
//            TAP_ExtractMjd(et[i].end_date, &year, &month, &day, &weekDay);
//            sprintf(&line[j],"%04d%02d%02d%02d%02d|", year, month, day, et[i].end_hour, et[i].end_min);
//want to export in local time
//need to convert et[i].end_date/time to local time
            utc = (et[i].end_date << 16) + (et[i].end_hour<<8) + et[i].end_min;
            local = exTAP_UTCtoLocalTime(utc);
            TAP_ExtractMjd(local>>16, &year, &month, &day, &weekDay);
            sprintf(&line[j],"%04d%02d%02d%02d%02d|", year, month, day, (local>>8)&0xff, local&0xff);

            j += strlen(&line[j]);
            // LCN of channel
            //need to map service_id to lcn
            lcn = get_channel_number(et[i].service_id);
            sprintf(&line[j],"%d|", lcn);
            j += strlen(&line[j]);
            // program title
            if (et[i].event_name) {
                memcpy(&line[j], et[i].event_name, et[i].event_name_length);
            }
            j += et[i].event_name_length;
            sprintf(&line[j],"|");
            j += strlen(&line[j]);
            // secondary title, e.g. episode title
            sprintf(&line[j],"|");      //episode
            j += strlen(&line[j]);
            // if this is a repeat
            sprintf(&line[j],"|");      //repeat
            j += strlen(&line[j]);
            // if this is first screening of a new show - not in RT data
            sprintf(&line[j],"|");      //new_series
            j += strlen(&line[j]);
            // production date
            sprintf(&line[j],"|");      //year;
            j += strlen(&line[j]);
            // if it is a premiere
            sprintf(&line[j],"|");      //premiere
            j += strlen(&line[j]);
            // last chance
            sprintf(&line[j],"|");      //last_chance
            j += strlen(&line[j]);
            // program description
            if (et[i].event_name) {
                sprintf(&line[j],"%s|", &et[i].event_name[et[i].event_name_length]);
            }
            else {
//v0.02
//                sprintf(&line[j],"");
                sprintf(&line[j],"|");
            }
            j += strlen(&line[j]);
            // program genre
            err = get_genre(&et[i], &genre);
            if (err == 0) {
                sprintf(&line[j],"%s", genre);
                j += strlen(genre);
            }
            sprintf(&line[j],"|");
            j += strlen(&line[j]);
            // program length
            sprintf(&line[j],"%d|", et[i].duration_hour*60+et[i].duration_min);
            j += strlen(&line[j]);
            // widescreen
            sprintf(&line[j],"|");      //widescreen
            j += strlen(&line[j]);
            // whether it is in colour
            sprintf(&line[j],"|");      //black_and_white
            j += strlen(&line[j]);
            // whether there are subtitles
            sprintf(&line[j],"|");      //subtitles
            j += strlen(&line[j]);
            // episode number
            sprintf(&line[j],"|");      //episode number      (<blank>,3/3)
            j += strlen(&line[j]);
            // episodes start at 1
            sprintf(&line[j],"|");      //episodes start at 1 (<blank>,xmltv_ns)
            j += strlen(&line[j]);
            // rating system
            sprintf(&line[j],"|");      //rating system       (<blank>,BBFC)
            j += strlen(&line[j]);
            // rating
            if (et[i].parental_rating) sprintf(&line[j],"%d|", et[i].parental_rating); else sprintf(&line[j],"|");
            j += strlen(&line[j]);
            // star rating
            sprintf(&line[j],"");       //star_rating
            j += strlen(&line[j]);

            sprintf(&line[k],"%06d", j-k-7);     //length, store at beginning of line,
            line[k+6] = '|';                     //        does not include length field
                                                 //        does not include newline

            line[j++] = 0x0d;                    //dos or unix?
            line[j++] = 0x0a;

            num_events++;
        }
    }

    //save to disk
    TAP_Print("\ni=%d, saving %x bytes ...", i, j);
    TAP_Hdd_Fwrite(line, j, 1, Op);

    TAP_Print("\nnum_events = %d", num_events);

    TAP_Hdd_Fclose(Op);
    free(line);

TAP_Print("\ndone");
    return 0;
}

//----------------------------------------------------------------------------

int ten_minute_update()
{
    int channel_type, channel_number;
    TYPE_TapChInfo channel_info;
    word mjd;
    byte hour, min, sec;

    TAP_Channel_GetCurrent(&channel_type, &channel_number);
    TAP_Channel_GetInfo(channel_type, channel_number, &channel_info);
    //see if are tuned to 4tv
    if (channel_info.svcId == 0x6980) {
        //see if it is time to wake up
        TAP_GetTime(&mjd, &hour, &min, &sec);
        if (dsmcc_hook_id < 0) {
            if (hour==3) {
                sprintf(strLog,"Applying DSMCC Hook");
                log_message(LOG_SCREEN_CONT|LOG_SERIAL, strLog, DELAY_SHORT);
                apply_dsmcc_hook();
            }
        }
        else
        //see if it is time to shut up
        if (dsmcc_hook_id >= 0) {
            if (hour==4) {
                sprintf(strLog,"Removing DSMCC Hook");
                log_message(LOG_SCREEN_CONT|LOG_SERIAL, strLog, DELAY_SHORT);
                remove_dsmcc_hook();
            }
        }
    }
}

//----------------------------------------------------------------------------
//
//print messages to serial port, screen, or file
//
int log_message(int log, char *str, int delay)
{
    char filename[] = "epg_extender.log";
    char logmsg_file[] = "\nlog: ";

    if (log & LOG_SERIAL) {
        TAP_Print("\n%s: %s", (char *)__tap_program_name__, str);
    }
    if (log & LOG_SCREEN_CONT) {
        ShowMessage((char *)__tap_program_name__, str, delay, 0);
    }
    if (log & LOG_SCREEN_WAIT) {
        ShowMessage((char *)__tap_program_name__, str, delay, 1);
    }
    if (log & LOG_FILE) {
        if (!OpLog) {
            TAP_Hdd_ChangeDir("/");
            TAP_Hdd_ChangeDir("ProgramFiles");
            //if exists delete it
            if (TAP_Hdd_Exist(filename)) {
                TAP_Hdd_Delete(filename);
            }
            TAP_Hdd_Create(filename, ATTR_NORMAL);
            OpLog = TAP_Hdd_Fopen(filename);
            if (!OpLog) {
                sprintf(strLog, "ERROR: cannot open logfile, %s", filename);
                log_message(LOG_SCREEN_WAIT|LOG_SERIAL, strLog, DELAY_SHORT);
                return -1;
            }
        }
        TAP_Hdd_Fwrite(logmsg_file, strlen(logmsg_file), 1, OpLog);
        TAP_Hdd_Fwrite(str, strlen(str), 1, OpLog);
    }
    return 0;
}
//----------------------------------------------------------------------------
